# stdlib imports
import argparse
import csv
import multiprocessing
import os
import re
import sys
import time
from collections import Counter
from functools import partial
# noinspection PyUnresolvedReferences
from types import ModuleType, FunctionType  # PyCharm can't 'see' ModuleType - hence unresolved reference
import builtins
import numpy
# rdkit imports
from rdkit import Chem
from rdkit.Chem import SaltRemover, Crippen, Fragments, rdMolDescriptors as rdMD, Descriptors as Desc
# MolVS
from molvs import Standardizer
# other optibrium imports
from qsb_classes import Assay, AssaySet, TestSet
from qsb_utils import load_pickle, save_pickle, create_directory, get_hash, time_diff, set_process_priority
from qsb_build_sets import build_model_sets
from qsb_test_sets import fit_and_test_models

# Raise exception on NumPy divide by zero or invalid operation rather than warning
numpy.seterr(divide='raise', invalid='raise')

__docstring__ = """A tool to clean raw ChEMBL data and guide users in compiling a clean set for QSAR modelling by 
                   building and testing models and from combinations of assays and reporting some helpful data"""
__author__ = 'Travis Hesketh (travis@optibrium.com)'
__copyright__ = 'Copyright (c) Optibrium Ltd. 2018'
__version__ = '2018.03.1'
__licence__ = 'GNU GPL v3'  # https://www.gnu.org/licenses/gpl-3.0.en.html

# <----- Configuration data -----> PEP8 says that GLOBAL VARIABLES SHOULD BE SHOUTED
CACHE_DIR = os.path.expanduser('~/.qsarsetbuilder')  # Hidden dirs less messy. See below for Windows
if not os.path.isdir(CACHE_DIR):
    try:
        assert create_directory(CACHE_DIR) == CACHE_DIR
    except AssertionError:
        raise OSError('Could not create cache directory, please create a directory at {}'.format(CACHE_DIR))
    if os.name == 'nt':  # Windows API call to hide the directory.
        from ctypes import wintypes, WinDLL
        kernel32 = WinDLL('kernel32')
        # Converting string and integer to C types explicitly
        kernel32.SetFileAttributesW(wintypes.LPCWSTR(CACHE_DIR), wintypes.DWORD(0x2))

# The field names needed in the document. Change the list as applicable.
SMILES, TYPE, P_VALUE, ASSAY_ID = 'CANONICAL_SMILES', 'STANDARD_TYPE', 'PCHEMBL_VALUE', 'ASSAY_CHEMBLID'
RELATION, S_VALUE, S_UNITS, DESCRIPTION = 'RELATION', 'STANDARD_VALUE', 'STANDARD_UNITS', 'DESCRIPTION'
VITAL_FIELDNAMES = SMILES, TYPE, P_VALUE, ASSAY_ID

PROPERTIES_TO_CACHE = ['IC50', 'EC50', 'ED50', 'AC50', 'Ki', 'Kd']  # Feel free to add to this
DESCRIPTOR_NAMES = ['MWgt', 'AlogP', 'TPSA', 'VABC', 'nAroRings', 'nRotBonds', 'HBAM', 'HBA', 'HBDM', 'HBD',
                    'NegC', 'PosC'] + [func[0] for func in Fragments.fns]  # Remember to change if changing descrs

# Using a dictionary over a namespace is for convenience: because ARGS is populated in a function, PyCharm can't
# 'see' the available fields and complains about unresolved references.
ARGS = dict()  # To be populated from command line arguments at runtime

#  <----- Caching ----->
PICKLED_DESCRIPTORS = os.path.join(CACHE_DIR, 'descriptors_dict.gz.pickle')
PICKLED_FINGERPRINTS = os.path.join(CACHE_DIR, 'fingerprint_dict.gz.pickle')
PICKLED_TRANSFORMATIONS = os.path.join(CACHE_DIR, 'transform_dict.gz.pickle')

try:
    DESCRIPTOR_DICT = load_pickle(PICKLED_DESCRIPTORS)
except FileNotFoundError:
    DESCRIPTOR_DICT = {}
D_DICT_HASH = hash(frozenset(DESCRIPTOR_DICT))
try:
    FINGERPRINT_DICT = load_pickle(PICKLED_FINGERPRINTS)
except FileNotFoundError:
    FINGERPRINT_DICT = {}
F_DICT_HASH = hash(frozenset(FINGERPRINT_DICT))
try:
    TRANSFORMATION_DICT = load_pickle(PICKLED_TRANSFORMATIONS)
except FileNotFoundError:
    TRANSFORMATION_DICT = {}
T_DICT_HASH = hash(frozenset(TRANSFORMATION_DICT))

# <----- Substrate searches ----->
CHEMLISTEM_MODEL_FILE = os.path.join(CACHE_DIR, 'minimodel_tf1.3.json')

# These are sometimes matched by chemlistem but we're probably not interested in them as substrates.
FALSE_MATCHES = {'amino', 'aminoacid', 'alkaloids', 'pharmacol', 'pubchem', 'concentration', 'compounds',
                 'fluorometer', 'phosphate', 'acetonitrile', 'nadph'}
# A dictionary of misspellings I've encountered in ChEMBL23. Feel free to add to this.
SUBSTRATE_MISTAKES = {'bezyl': 'benzyl', 'quiniline': 'quinoline'}

# Fairly harsh matching to ensure most differences in chemical name formatting are caught. Unfortunately,
# if regioisomers are substrates (2-methyl vs 4-methyl etc.) they will be treated as the same.
SUBSTRATE_PATTERN = r"[-,]*[NO][-,]|\[3H]|\[14C]|\[35S]|\[125I]|[0-9{})([\-\],'\s]|beta|alpha|substrate"
SUBSTRATE_NAME_REGEX = re.compile(SUBSTRATE_PATTERN, flags=re.IGNORECASE)

# To remove Python's brackets and inverted commas in string representations of lists, sets and dicts.
FORMATTING_REGEX = re.compile(r"set\(|[(){}\[\]']")


def update_cache(compound_data):
    """Updates the global dictionaries containing the SMILES: descriptor and fingerprint values, and the input SMILES
    to canonical SMILES dictionary. These are saved to files in CACHE_DIR as gzipped Python pickles.
    """
    global DESCRIPTOR_DICT, FINGERPRINT_DICT, TRANSFORMATION_DICT

    for smiles, descriptor, fingerprint, i_smiles in compound_data:
        if descriptor is not None:
            DESCRIPTOR_DICT[smiles] = descriptor
        if fingerprint is not None:
            FINGERPRINT_DICT[smiles] = fingerprint
        if i_smiles is not None:
            TRANSFORMATION_DICT[i_smiles] = smiles

    if ARGS['no_cache']:  # Still need the globals to be updated, but don't want to overwrite the cache
        return

    if hash(frozenset(DESCRIPTOR_DICT)) != D_DICT_HASH:
        save_pickle(DESCRIPTOR_DICT, PICKLED_DESCRIPTORS)
    if hash(frozenset(FINGERPRINT_DICT)) != F_DICT_HASH:
        save_pickle(FINGERPRINT_DICT, PICKLED_FINGERPRINTS)
    if hash(frozenset(TRANSFORMATION_DICT)) != T_DICT_HASH:
        save_pickle(TRANSFORMATION_DICT, PICKLED_TRANSFORMATIONS)


# <----- Core Functions ----->
def process_args(arguments=None):
    """Populates global ARGS dict from the command line or from a list of arguments.
    """
    global ARGS

    arg_parser = argparse.ArgumentParser(
        description='{}. Author: {}'.format(__docstring__, __author__),
        prog='qsarsetbuilder',
        epilog='{}. Licence: {}'.format(__copyright__, __licence__))
    arg_parser.add_argument('files',
                            help='file names to be analyzed', type=str, nargs='*')
    arg_parser.add_argument('-t', '--test-set',
                            help='path to external test set',
                            type=str, default=None, metavar='file')
    arg_parser.add_argument('-m', '--models',
                            help='types of model to build, where c is classification and r is regression. Default r',
                            type=str, choices=['c', 'r', 'cr', 'rc'], default='r')
    arg_parser.add_argument('-c', '--num-cores',
                            help='number of CPU cores to use. Default all',
                            type=int, default=os.cpu_count(), metavar='n')
    arg_parser.add_argument('-s', '--split',
                            help='split for classification models (as pX value). Default 6 (e.g. 1uM)',
                            type=int, default=6, metavar='n')
    arg_parser.add_argument('-n', '--num-sets',
                            help="""number of sets to generate and test. If this is larger than the number of possible 
                                    combinations, sets are built combinatorially instead. Default 10,000.""",
                            type=int, default=10000, metavar='n')
    arg_parser.add_argument('-o', '--cont-after',
                            help="""The run is terminated if no good sets are found in this many runs. As sets are
                                    tested in order of increasing size, this means that larger sets are left out when
                                    their performance starts to become poor. Default 1000""",
                            type=int, default=1000, metavar='n')
    arg_parser.add_argument('-l', '--max-length',
                            help="""Max size of assay set (as number of compounds). Set to 0 for no limit. This is
                                    ignored if assay sets are being built combinatorially. Default 2000""",
                            type=int, default=2000, metavar='n')
    arg_parser.add_argument('-p', '--properties',
                            help="""property (or properties) to predict. Default option includes IC50, EC50,
                                    ED50 and AC50, Ki and Kd. Special options: ALL tries to build a model from all
                                    default types, XC50 tries IC50, EC50, ED50 and AC50 together""",
                            default=PROPERTIES_TO_CACHE, nargs='*', metavar='property')
    arg_parser.add_argument('--encoding',
                            help="""encoding to use for TSV/CSV files. For StarDrop, try windows-1252. For more options,
                                    consult the online python documentation: 
                                        https://docs.python.org/3/library/codecs.html#standard-encodings
                                    Default utf-8""",
                            type=str, default='utf-8', metavar='encoding')
    arg_parser.add_argument('--priority',
                            help="""priority of the processes spawned during multiprocessing. On unix systems, values 
                                    higher than 'normal' do nothing (requires root access). Default below_normal""",
                            type=str, choices=['idle', 'below_normal', 'normal', 'above_normal', 'high'],
                            default='below_normal')
    arg_parser.add_argument('--verbosity',
                            help='level of verbosity. (0-2). 0 is no terminal output, 2 is full output. Default 1',
                            type=int, metavar='n', default=1)
    arg_parser.add_argument('--excl-terms',
                            help="""terms to exclude. If any of these are in the description, the assay will not be 
                                    included. These are not case sensitive. Leave blank for no exclusions. Defaults are 
                                    some terms which refer to HTS studies, insect cells and oocytes""",
                            default=['high throughput', 'insect', 'baculovirus', 'oocytes'], nargs='*', metavar='term')
    arg_parser.add_argument('--excl-assays',
                            help='assays to exclude (by assay id)',
                            nargs='*', metavar='assay', default=list())
    arg_parser.add_argument('--substrates',
                            help="""use chemlistem to analyse descriptions and run by chemical name. This could take a
                                    long time""",
                            action='store_true'),
    arg_parser.add_argument('--no-oor',
                            help="""Don't use values which have their associated -log(M) value missing. In ChEMBL, This 
                                    usually means that the data is considered 'out of range' (pX <4)""",
                            action='store_false')
    arg_parser.add_argument('--no-cache',
                            help='don\'t use any cached data for this run (including file-specific)',
                            action='store_true')
    arg_parser.add_argument('--purge-cache',
                            help='clear the caches for descriptors, fingerprints and transformations',
                            action='store_true')

    args = arg_parser.parse_args(arguments)
    ARGS = vars(args)


def validate_input(input_file_path):
    """Takes input files, expands/checks their path and returns the necessary information: the file path, delimiter
    and 'model name'
    """
    input_directory, input_file_name = os.path.split(input_file_path)

    if not input_directory:  # If there's no path provided, assume file is in the working directory.
        input_directory = os.getcwd()
    elif '~' in input_directory:  # Need to expand user dir
        input_directory = os.path.expanduser(input_directory)
    input_file_path = os.path.join(input_directory, input_file_name)

    if not os.path.isfile(input_file_path):
        raise IOError(input_file_path + ' is not a file.')

    model_name, extension = os.path.splitext(input_file_name)
    delimiter = ',' if extension.lower() == '.csv' else '\t'
    return input_file_path, delimiter, model_name


def init_pool(args, d_dict, f_dict, t_dict=None, vars_to_keep=None):
    """Initializer for multiprocessing Pools.
    Sets process priority and sets global variables in the child process. On POSIX, the latter shouldn't make much of a
    difference, but is necessary on Windows to have access to global variables which are populated within a function
    (e.g. ARGS) because it does not implement 'forking'
    """
    # Editing the globals() dict for each process (dict of variable_name: object for everything in the global namespace)
    # changes the global namespace in that process
    process_global_dict = globals()
    process_global_scope = list(process_global_dict.keys())  # List of variable names in global scope as strings.
    kept_types = [ModuleType, FunctionType, type, str]  # Types we always want to keep. type == type(class)
    builtin_objects = vars(builtins).keys()  # The names of the builtin functions and types as a list of strings

    for variable in process_global_scope:
        if variable in builtin_objects \
                or type(process_global_dict[variable]) in kept_types \
                or (vars_to_keep and variable in vars_to_keep):
            continue
        del process_global_dict[variable]

    # Setting those which we want
    global ARGS, DESCRIPTOR_DICT, FINGERPRINT_DICT, TRANSFORMATION_DICT
    ARGS, DESCRIPTOR_DICT, FINGERPRINT_DICT = args, d_dict, f_dict
    TRANSFORMATION_DICT = t_dict if t_dict is not None else dict()
    # This is useful to avoid system becoming unresponsive if using all cores.
    set_process_priority(ARGS['priority'])


def canonicalize_smiles(smiles):  # American spelling is the standard for function names, sadly...
    """Takes a smiles string and returns a standardised representation. This function strips stereochemistry
       and salts (as we're using 2D descriptors, neither are necessary) and returns the canonical tautomer.
    """
    mol = Chem.MolFromSmiles(smiles)
    if not mol or mol.GetNumAtoms() is 0:
        return
    Chem.RemoveStereochemistry(mol)  # None of our descriptors take stereochemistry into account.
    mol = SaltRemover.SaltRemover().StripMol(mol, dontRemoveEverything=True)
    # MolVS is clever. Tautomers and functional groups are returned as standardised representations.
    mol = Standardizer().tautomer_parent(mol)
    return Chem.MolToSmiles(mol)


def calc_descrs_fingerprint(smiles):
    """A function to calculate descriptor/fingerprint values directly from a canonical smiles string.
       This is used if a cache file is read in and the descriptors and fingerprints are missing.
    """
    mol = Chem.AddHs(Chem.MolFromSmiles(smiles))  # If we're using this function, assume canonical.
    return smiles, calculate_descriptors(mol), rdMD.GetMorganFingerprint(mol, 2), None


def validate_data(row, external=False):
    """Validates data in row, and calculates descriptors and fingerprints. This function is used to map the read-in
       rows to the process pool and as such is run on multiple threads.
    """
    if not external:  # The external set probably only has SMILES and P_VALUE.
        if not row[SMILES] \
                or row[TYPE] not in PROPERTIES_TO_CACHE \
                or (RELATION in row and row[RELATION] != '=') \
                or (not row[P_VALUE] and not (row.get(S_UNITS) and row.get(S_VALUE))):
            return

    # Validation of the numerical value. Try for the log-shifted value. If it's missing, calculate the value from the
    # 'standard value' if allowed for by command line args (i.e. not --no-oor)
    try:
        value = float(row[P_VALUE])
    except ValueError:
        if ARGS['no_oor']:  # If we're not using ChEMBL's out of range values, return here
            return

        unit_exponents = dict(nm=-9, um=-6, mm=-3, m=0)
        try:
            units, std_value = row[S_UNITS], row[S_VALUE]
            power = unit_exponents[units]
            std_value_m = float(std_value) * (10 ** power)
            value = round(-numpy.log10(std_value_m), 2)
        except (KeyError, ValueError, FloatingPointError):
            return
        else:
            if not numpy.isfinite(value) or value <= 0:
                return

    i_smiles = row[SMILES]  # isomeric, non-standard SMILES
    smiles = TRANSFORMATION_DICT[i_smiles] if i_smiles in TRANSFORMATION_DICT else canonicalize_smiles(i_smiles)
    if not smiles:
        return
    row[P_VALUE], row[SMILES] = value, smiles

    # Calculating our missing descriptor and fingerprint values.
    mol = Chem.AddHs(Chem.MolFromSmiles(smiles))
    descriptor = DESCRIPTOR_DICT[smiles] if smiles in DESCRIPTOR_DICT else calculate_descriptors(mol)
    fingerprint = FINGERPRINT_DICT[smiles] if smiles in FINGERPRINT_DICT else rdMD.GetMorganFingerprint(mol, 2)

    assay_id, assay_type, description = [row.get(x, '') for x in (ASSAY_ID, TYPE, DESCRIPTION)]
    return row, (smiles, descriptor, fingerprint, i_smiles), (assay_id, assay_type, smiles, value, description)


def read_in_assays(input_file, delimiter, output_file, cached_file):
    """Reads the input file, calculating descriptor, fingerprint and transformation values using multiple processes,
       and caches it.
    """
    vars_to_keep = ['DESCRIPTOR_DICT', 'FINGERPRINT_DICT', 'TRANSFORMATION_DICT', 'PROPERTIES_TO_CACHE']
    init_args = [ARGS, DESCRIPTOR_DICT, FINGERPRINT_DICT, TRANSFORMATION_DICT, vars_to_keep]
    mp_pool = None  # Lazy load. Pool initialization might not be necessary

    output, output_fieldnames, compound_data, assay_data = list(), list(), list(), list()
    n_data = int()
    # Using SHA1 hash so that cache isn't loaded if the file has changed
    file_hash = get_hash(input_file)

    # If the input file has already been cached, read in from that...
    if os.path.exists(cached_file) and not ARGS['no_cache']:
        cached_hash, cached_data = load_pickle(cached_file)
        if cached_hash == file_hash or cached_file == input_file:  # ...but only if the input file is the same
            output, output_fieldnames, n_data, assay_data = cached_data

            # If the descriptor/fingerprint cache has been purged in the meantime, need to re-calculate them.
            # Stored SMILES strings are canonical, so the transformations will still be missing.
            compounds = {datapoint[2] for datapoint in assay_data}
            if not compounds <= DESCRIPTOR_DICT.keys() or not compounds <= FINGERPRINT_DICT.keys():
                if mp_pool is None:
                    mp_pool = multiprocessing.Pool(ARGS['num_cores'], initializer=init_pool, initargs=init_args)
                compound_data = list(mp_pool.imap_unordered(calc_descrs_fingerprint, compounds))
                update_cache(compound_data)

    if ARGS['test_set']:
        validate_ext = partial(validate_data, external=True)
        test_set_path, test_set_delimiter, _ = validate_input(ARGS['test_set'])

        with open(test_set_path, 'r', newline='', encoding=ARGS['encoding']) as e_test_set_object:
            reader = csv.DictReader(e_test_set_object, delimiter=test_set_delimiter)

            test_set_compound_data, test_set_data = list(), list()
            if mp_pool is None:
                multiprocessing.Pool(ARGS['num_cores'], initializer=init_pool, initargs=init_args)
            for result in mp_pool.imap_unordered(validate_ext, reader, chunksize=4):
                if result is not None:
                    test_set_compound_data.append(result[1])
                    test_set_data.append(result[2])
            update_cache(test_set_compound_data)
        external_test_set = TestSet(test_set_data, DESCRIPTOR_DICT)
    else:
        external_test_set = None

    if not output:  # No cache; reading in from file.
        with open(input_file, 'r', newline='', encoding=ARGS['encoding']) as input_file_object:
            reader = csv.DictReader(input_file_object, delimiter=delimiter)

            output_fieldnames = reader.fieldnames
            for fieldname in VITAL_FIELDNAMES:
                if fieldname not in output_fieldnames:
                    raise IOError('{0}: Failed read in, necessary field {1} not in file.'.format(input_file, fieldname))

            if not mp_pool:
                mp_pool = multiprocessing.Pool(ARGS['num_cores'], initializer=init_pool, initargs=init_args)
            for result in mp_pool.imap(validate_data, reader, chunksize=4):
                if result is not None:
                    output.append(result[0])
                    compound_data.append(result[1])
                    assay_data.append(result[2])
            update_cache(compound_data)

            n_data = len(output)
            cached_data = output, output_fieldnames, n_data, assay_data
            save_pickle((file_hash, cached_data), cached_file)

    # Setting up the Assay classes
    assays = dict()
    for assay_id, assay_type, smiles, value, description in assay_data:
        assay = assays[assay_id] if assay_id in assays else Assay(assay_id, assay_type)
        assay.description = description
        assay[smiles] = value
        assays[assay_id] = assay

    # Removing assays with less than three compounds and those which are subsets of others.
    assay_blacklist, finalized_assays, unique_compounds = set(), set(), set()
    for assay_id_x, assay_x in assays.items():
        if len(assay_x) < 3 or assay_id_x in assay_blacklist:
            assay_blacklist.add(assay_id_x)
            continue

        for assay_id_y, assay_y in assays.items():
            if assay_id_x == assay_id_y or assay_id_y in assay_blacklist:
                continue
            elif assay_x > assay_y or len(assay_y) < 3:
                assay_blacklist.add(assay_id_y)
                continue
            elif assay_x < assay_y or (assay_x == assay_y and assay_id_y in finalized_assays):
                assay_blacklist.add(assay_id_x)
                break
        else:
            unique_compounds |= assay_x
            finalized_assays.add(assay_id_x)

    for assay_id in assay_blacklist:
        del assays[assay_id]

    # Writing cleaned data out to file.
    sorted_output = sorted(output, key=lambda field: (field[ASSAY_ID], field[P_VALUE]))
    with open(output_file, 'w', newline='', encoding=ARGS['encoding']) as output_file_object:
        csv_writer = csv.DictWriter(output_file_object, delimiter=',', fieldnames=output_fieldnames)
        csv_writer.writeheader()
        for row in sorted_output:
            if row.get(ASSAY_ID) not in assay_blacklist:
                csv_writer.writerow(row)

    return assays, external_test_set, len(unique_compounds), n_data


def build_property_list():
    """Builds a list of the properties to include in the run from the command line options.
    """
    property_list = list()

    for property_name in ARGS['properties']:
        if property_name == 'ALL':
            properties = set(PROPERTIES_TO_CACHE)
        elif property_name == 'XC50':
            properties = {x for x in PROPERTIES_TO_CACHE if x.endswith('50')}
        else:
            properties = {property_name}
        property_list.append((property_name, properties))

    return property_list


def select_assays(assays, accepted_properties, substrate):
    """Selects Assay classes to be used in this run, eliminating assays which are subsets of others and those which
       don't meet the criteria (excluded assays or terms)
    """
    unique_compounds = set()
    n_used_data = int()
    assays_to_run, finalized_assays = dict(), dict()

    for assay_id, assay in assays.items():
        contains_excluded_terms = True in [term in assay.description.lower() for term in ARGS['excl_terms']]

        if assay.assay_type in accepted_properties \
                and (substrate == 'all' or substrate in assay.substrates) \
                and assay_id not in ARGS['excl_assays'] \
                and not contains_excluded_terms:
            assays_to_run[assay_id] = assay

    # Searching for assays which are exact subsets of others (i.e, papers which repeat old results)
    for assay_id_x, assay_x in assays_to_run.items():
        assay_x.overlaps = dict()
        for assay_id_y, assay_y in finalized_assays.items():
            if assay_x & assay_y:
                square_distances = list()
                for compound in assay_x & assay_y:
                    square_distances.append((assay_x[compound] - assay_y[compound]) ** 2)
                rmsd = numpy.sqrt(numpy.mean(square_distances))
                assay_x.update_overlaps(assay_y, rmsd)

        finalized_assays[assay_id_x] = assay_x
        unique_compounds |= assay_x
        n_used_data += len(assay_x)

    return finalized_assays, n_used_data, len(unique_compounds)


def find_substrates(assays):
    """Analyse the assay descriptions for potential probe substrates using chemlistem and save these to the classes.
       Downloads the chemlistem minimal model using urllib if not present.
    """
    std_substrates = list()

    if ARGS['verbosity'] == 0:  # Redirecting stdout to nowhere - chemlistem (tensorflow) is noisy
        sys.stdout, sys.stderr = open(os.devnull, 'w'), open(os.devnull, 'w')
    sys.stdout.write('\nThe following warning messages pertain to Chemlistem. This should take around 30s to load.\n'
                     + '----------------------------------------------------------------------------------------\n')
    try:
        from chemlistem import minimodel
    except ImportError:
        # noinspection PyUnusedLocal
        minimodel = None
        sys.stdout.write('chemlistem and h5py are required to analyse substrates. These can be installed using pip.\n')
        sys.stdout, sys.stderr = sys.__stdout__, sys.__stderr__
        return std_substrates

    model = minimodel.MiniModel()
    if not os.path.exists(CHEMLISTEM_MODEL_FILE):
        from urllib.request import urlretrieve
        urlretrieve('https://bitbucket.org/rscapplications/chemlistem/downloads/minimodel_tf1.3.h5',
                    CHEMLISTEM_MODEL_FILE)  # Chemlistem's 'get_mini_model()' doesn't seem to work with tf 1.3
    model.load(CHEMLISTEM_MODEL_FILE)

    time.sleep(0.25)  # Just to ensure that the next line is printed after the chemlistem message.
    sys.stdout.write('-----------------------------------------------------------------------------------------\n'
                     + 'Finished read in for Chemlistem. Scanning for substrates.\n\n')
    if ARGS['verbosity'] == 0:
        sys.stdout, sys.stderr = sys.__stdout__, sys.__stderr__

    substrate_set = set()
    for assay in assays.values():
        for mistake, correction in SUBSTRATE_MISTAKES.items():
            if mistake in assay.description:
                assay.description = assay.description.replace(mistake, correction)

        results = model.process(assay.description)
        if results:
            for result in results:
                chemical_name = SUBSTRATE_NAME_REGEX.sub('', result[2]).lower()
                if chemical_name not in FALSE_MATCHES and len(chemical_name) > 4:
                    substrate_set.add(chemical_name)

    for substrate_x in list(substrate_set):
        for substrate_y in substrate_set:
            if substrate_x != substrate_y and substrate_y in substrate_x:
                break
        else:
            continue
        substrate_set.remove(substrate_x)

    substrate_counts = Counter()
    for assay in assays.values():
        for substrate in substrate_set:
            if substrate in SUBSTRATE_NAME_REGEX.sub('', assay.description).lower():
                assay.substrates.add(substrate)
                substrate_counts[substrate] += len(assay)

    substrates = [substrate for substrate, count in substrate_counts.items() if count > 150]
    substrates = sorted(substrates, key=lambda sub: substrate_counts[sub], reverse=True)
    return substrates


def calculate_descriptors(mol):
    """Takes RDKit molecule as input, calculates descriptors and returns them as a list. If changing this,
       remember to update DESCRIPTOR_NAMES
    """
    def calc_vabc(molecule):
        """Takes RDKit molecule as input, calculates and returns the VABC descriptor, an approximation of van der Waals
           volume. Method from 10.1021/jo034808o
        """
        coefficients = dict(C=20.58, H=7.24, N=15.6, O=14.71, Cl=22.45, Br=26.52, F=13.31, I=32.52, S=24.43, P=24.43,
                            As=26.52, Se=28.73, Si=38.79)
        atom_contributions = [coefficients.get(atom.GetSymbol(), 0) for atom in molecule.GetAtoms()]
        num_atoms, total_contribution = len(atom_contributions), sum(atom_contributions)
        num_aro_rings, num_ali_rings = rdMD.CalcNumAromaticRings(molecule), rdMD.CalcNumAliphaticRings(molecule)
        num_atoms_and_rings = num_atoms + num_aro_rings + num_ali_rings

        vdw_vol = total_contribution - (num_atoms_and_rings - 1) * 5.92 - num_aro_rings * 14.7 - num_ali_rings * 3.8
        return round(vdw_vol, 1)

    wm_descriptors = [Desc.MolWt(mol),
                      Crippen.MolLogP(mol),
                      rdMD.CalcTPSA(mol),
                      calc_vabc(mol),
                      rdMD.CalcNumAromaticRings(mol)] + rdMD.MQNs_(mol)[18:25]
    topological_matches = [func[1](mol) for func in Fragments.fns]
    return wm_descriptors + topological_matches


def test_datasets(external_test_set, assays, assay_set_ids):
    """Runs and tests the assay sets generated by build_model_sets. Rebuilds AssaySet classes from a list of IDs,
       fills in the descriptors and fingerprints, tests them and returns them for analysis.
    """
    set_id = assay_set_ids.pop()
    assay_set = AssaySet(set_id)
    assay_set.add_assays(*[assays[x] for x in assay_set_ids])

    for compound, value in assay_set:
        assay_set.descriptors[compound] = DESCRIPTOR_DICT[compound]
        assay_set.fingerprints[compound] = FINGERPRINT_DICT[compound]
    assay_set = fit_and_test_models(assay_set, ARGS['split'], ARGS['models'], external_test_set)
    return assay_set


def produce_report(counters, arrays, assays, n_good_sets, report_header, model_directory):
    """Produces the report, assay similarity matrix and set distribution histogram from the model information,
       outputting them to the model directory.
    """
    from matplotlib import pyplot as plt
    plt.switch_backend('agg')

    report_file = os.path.join(model_directory, 'report.txt')
    distribution_file = os.path.join(model_directory, 'set_size_distrib.png')
    distance_matrix_file = os.path.join(model_directory, 'assay_similarity_array.csv')

    assay_list = sorted(assays.keys())
    assay_counter, good_assay_counter, length_counter, r_descr_counter, c_descr_counter = counters
    coappearance_array, good_coappearance_array = arrays

    for row_index, row in enumerate(good_coappearance_array):
        for column_index, value in enumerate(row):
            total_value = coappearance_array[row_index, column_index]
            good_coappearance_array[row_index, column_index] = 0 if value == 0 else (value / total_value)

    with open(distance_matrix_file, 'w', encoding=ARGS['encoding'], newline='') as distance_matrix_file_obj:
        csv_writer = csv.writer(distance_matrix_file_obj, delimiter=',')
        csv_writer.writerow([''] + assay_list)
        for assay_id, distances in zip(assay_list, good_coappearance_array):
            csv_writer.writerow([assay_id] + distances.tolist())

    with open(report_file, 'w') as report_file_obj:
        report_file_obj.write(report_header)
        report_file_obj.write('Assays appearing in successful models and their occurrence:\n')
        for assay_id, occurrence in sorted(good_assay_counter.items(), key=lambda count: count[1], reverse=True):
            report_file_obj.write('{}, {:0<4.2}: {}\n'.format(assay_id, occurrence / assay_counter[assay_id],
                                                              assays[assay_id].description))

        non_appearing_assays = assay_counter.keys() - good_assay_counter.keys()
        if non_appearing_assays:
            report_file_obj.write('\n\n')
            report_file_obj.write('Assays which did not appear in in successful models:\n')
            for assay_id in non_appearing_assays:
                report_file_obj.write('{} (appeared in {} sets)'.format(assay_id, assay_counter[assay_id]))

        if 'r' in ARGS['models']:
            report_file_obj.write('\n\n')
            report_file_obj.write('Descriptors appearing in successful regression models and their occurrence:\n')
            for name, occurrence in sorted(r_descr_counter.items(), key=lambda count: count[1], reverse=True):
                report_file_obj.write('{}, {:0<4.2}\n'.format(name, occurrence / n_good_sets))

        if 'c' in ARGS['models']:
            report_file_obj.write('\n\n')
            report_file_obj.write('Descriptors appearing in successful classification models and their occurrence:\n')
            for name, occurrence in sorted(c_descr_counter.items(), key=lambda count: count[1], reverse=True):
                report_file_obj.write('{}, {:0<4.2}\n'.format(name, occurrence / n_good_sets))

    lengths, counts = list(length_counter.keys()), list(length_counter.values())
    plt.hist(lengths, weights=counts, bins=20)
    plt.title('Set Size Distribution')
    plt.xlabel('N Compounds In Set (Set Size)')
    plt.ylabel('Number of Sets of Size N')
    plt.savefig(distribution_file)
    plt.clf()


def run(input_file_path):
    """Runs QSARSetBuilder for a given input file (expected input: ChEMBL data for a single target).
    Outputs a directory (modelname_analysis) containing:
        A cleaned version of the input file. All assays with < 3 compounds are removed, all qualified (~, <, >, etc.)
          values removed and standardised 2D representations of the SMILES string are returned. (CSV)
        A cached binary version of the information contained in the clean file. (.CBSRT)
        A directory (ISO time string e.g. 20180309T105903) containing run information:
            The run log file (.LOG)
            A directory for each substrate/property combination (e.g. midazolam_IC50):
                Report file (.TXT)
                Assay similarity matrix (.CSV)
                Set size distribution histogram (.PNG)
    """
    # Pretty and ISO-standard representations of the time string.
    initial_time = time.time()
    utc_time = time.gmtime(initial_time)
    string_time, iso_time = (time.strftime(x, utc_time) for x in ('%H:%M:%S - %d %b %Y UTC', '%Y%m%dT%H%M%SZ'))

    # Setting up directories and a few output files
    input_file, delimiter, model_name = validate_input(input_file_path)
    input_directory = os.path.split(input_file)[0]
    if input_file.endswith('.cbsrt'):  # If reading in the cache, don't make a subdirectory
        parent_directory = input_directory
    else:
        parent_directory = os.path.join(input_directory, (model_name + '_analysis'))
    parent_directory = create_directory(parent_directory)
    working_directory = create_directory(os.path.join(parent_directory, iso_time))
    output_file = os.path.join(parent_directory, (model_name + '_clean.csv'))
    cached_file = os.path.join(parent_directory, (model_name + '.cbsrt'))

    # Output streams - log file and terminal.
    log_file = open(os.path.join(working_directory, (model_name + '.log')), 'w', 1)
    terminal = sys.stdout if ARGS['verbosity'] != 0 else open(os.devnull, 'w')
    def write_out_line(line): terminal.write(line), log_file.write(line)

    # Starting the actual run now
    write_out_line(('Beginning read in for {0}: started at {1}\nCalculating descriptors and fingerprints, this could'
                    + ' take some time.\n').format(model_name, string_time))

    # Reading in, canonicalising smiles strings, ensuring all values are present and writing out the 'clean' data.
    all_assays, external_test_set, n_compounds, n_data = read_in_assays(input_file, delimiter, output_file, cached_file)

    write_out_line(('Read in and fingerprint/descriptor calculation complete. {} compounds ({} values) in {} assays.'
                    + '\nCleaned input file saved to: {}\n').format(n_compounds, n_data, len(all_assays), output_file))

    # Substrate analysis, if the command line flag is set
    std_substrates = find_substrates(all_assays) if ARGS['substrates'] else list()
    if std_substrates:
        write_out_line('Standard substrates found: {}\n'.format(std_substrates))
    std_substrates.append('all')

    # Checking which properties should be included
    property_list = build_property_list()

    # Building sets and running models for each substrate/property combination.
    for substrate in std_substrates:
        for property_name, accepted_properties in property_list:
            # Selecting assays which match the combination
            assays, n_used, n_unique = select_assays(all_assays, accepted_properties, substrate)
            # Maximum generated set size (in compounds): as long as possible unless a max is specified.
            max_len = n_unique if ARGS['max_length'] == 0 else min(n_unique, ARGS['max_length'])
            run_names = {'property': property_name, 'model': model_name, 'substrate': substrate}  # To format strings

            if not assays or n_unique < 150:  # Not enough to QSAR model.
                write_out_line('Not enough compounds for {property} (substrate: {substrate})\n'.format(**run_names))
                continue
            # If we have enough to model
            write_out_line(('Modelling {property} for {model} started. (substrate: {substrate})\n{0} compounds in {1} '
                            + 'assays use this substrate and type and are suitable for continuous modelling.\n')
                           .format(n_unique, len(assays), n_used, **run_names))

            # Building the sets. assay_set_gen is a generator which produces lists of assay_ids and which have the set
            # number appended (as a string).
            assay_set_gen, num_sets = build_model_sets(assays, ARGS['num_sets'], max_len)
            write_out_line('Set generation complete. {0} sets generated.\n'.format(num_sets))

            # Creating directory if we have sets to test.
            directory_name = '{substrate:.10}_{property}'.format(**run_names)
            model_directory = create_directory(os.path.join(working_directory, directory_name))
            set_list_file = os.path.join(model_directory, 'sets_list.csv')

            # Counters and arrays for the report
            index_dict = {assay_id: index for index, assay_id in enumerate(sorted(assays.keys()))}  # order for arrays
            coappearance_array = numpy.zeros((len(assays), len(assays)))
            good_coappearance_array = numpy.zeros((len(assays), len(assays)))
            arrays = (coappearance_array, good_coappearance_array)

            assay_counter, good_assay_counter = Counter(), Counter()
            r_descr_counter, c_descr_counter, length_counter = Counter(), Counter(), Counter()
            counters = (assay_counter, good_assay_counter, length_counter, r_descr_counter, c_descr_counter)
            # These int counters used for early termination and to keep track of number of sets/good sets
            last_tested, last_reset, n_good_sets = int(), int(), int()

            # Set list file line is written after set is tested.
            set_list_file_object = open(set_list_file, 'w', encoding=ARGS['encoding'], newline='')
            set_list_writer = csv.writer(set_list_file_object, delimiter=',')

            # Setting up the multiprocessing Pool.
            vars_to_keep = ['DESCRIPTOR_DICT', 'FINGERPRINT_DICT']
            init_args = [ARGS, DESCRIPTOR_DICT, FINGERPRINT_DICT, None, vars_to_keep]
            mp_pool = multiprocessing.Pool(ARGS['num_cores'], initializer=init_pool, initargs=init_args)
            test_dataset = partial(test_datasets, external_test_set, assays)  # Partial function to map to the pool

            for assay_set in mp_pool.imap(test_dataset, assay_set_gen, chunksize=4):
                set_list_writer.writerow([assay_set.id] + list(assay_set.ids))
                last_tested = int(assay_set.id)

                if assay_set.good_set:
                    write_out_line(assay_set.results)
                    n_good_sets, last_reset = n_good_sets + 1, int()

                    length_counter[len(assay_set)] += 1
                    for index in assay_set.r_descr_mask:
                        r_descr_counter[DESCRIPTOR_NAMES[index]] += 1
                    for index in assay_set.c_descr_mask:
                        c_descr_counter[DESCRIPTOR_NAMES[index]] += 1
                else:
                    last_reset += 1
                    if ARGS['verbosity'] < 2:
                        log_file.write(assay_set.results)
                    else:
                        write_out_line(assay_set.results)

                # Adding co-appearing assays to the arrays and counting the assays.
                for assay_id_x in assay_set.ids:
                    if assay_set.good_set:
                        good_assay_counter[assay_id_x] += 1
                    assay_counter[assay_id_x] += 1
                    for assay_id_y in assay_set.ids:
                        if assay_id_x == assay_id_y:
                            continue
                        if assay_set.good_set:
                            good_coappearance_array[index_dict[assay_id_x], index_dict[assay_id_y]] += 1
                        coappearance_array[index_dict[assay_id_x], index_dict[assay_id_y]] += 1

                # Early termination flag
                if ARGS['cont_after'] and last_reset == ARGS['cont_after']:
                    write_out_line('No good sets found in last {}, exiting run.\n'.format(ARGS['cont_after']))
                    break

            if good_coappearance_array.any():
                header = ('Report for models of {model}. Substrate: {substrate} Property name: {property} \n{0} '
                          + 'successful sets out of {1}\n\n').format(n_good_sets, last_tested, **run_names)
                produce_report(counters, arrays, assays, n_good_sets, header, model_directory)

    h, m, s = time_diff(initial_time, time.time())
    write_out_line('Complete. Total time taken for {0}: {1}h{2}m{3}s\n'.format(model_name, h, m, s))


if __name__ == '__main__':
    process_args()

    if not ARGS['files']:
        raise IOError('No files entered.')

    if ARGS['purge_cache'] or ARGS['no_cache']:
        DESCRIPTOR_DICT, FINGERPRINT_DICT, TRANSFORMATION_DICT = dict(), dict(), dict()
        D_DICT_HASH, F_DICT_HASH, T_DICT_HASH = int(), int(), int()

    for target in ARGS['files']:
        run(target)
