# stdlib imports
import tempfile
import json
from bisect import bisect_left
from collections import Counter
from copy import deepcopy
from random import choice, random, randint
# optibrium
from qsb_classes import AssaySet

__docstring__ = """Set building for QSARSetBuilder, a program to help build a clean data set for QSAR modelling from raw 
                   public data. Designed to work with ChEMBL data.
                """
__author__ = 'Travis Hesketh (travis@optibrium.com)'
__copyright__ = 'Copyright (c) Optibrium Ltd 2018. All rights reserved.'
__version__ = '2018.03.1'
__licence__ = 'GNU GPL v3'  # https://www.gnu.org/licenses/gpl-3.0.en.html

MEBIBYTE = 2 ** 20  # 1 MiB

NUM_SETS, MAX_LEN = int(), int()
# See docstring for add_to_global for notes on data structure in memory / on disk.
ID_COMBINATIONS = tempfile.SpooledTemporaryFile(256 * MEBIBYTE)
TEMP_ID_COMBINATIONS = tempfile.SpooledTemporaryFile(256 * MEBIBYTE)
LENGTH_BYTE_COUNTS = Counter()  # len(assay_set) (int) : number of bytes in ID_COMBINATIONS at that length (int)


def build_sets(assays):
    """Builds the initial sets based on RMSD of overlapping compounds.
    """
    assay_sets = list()

    for assay_id_x, assay_x in assays.items():
        if not assay_x.overlaps:
            continue

        assay_set = AssaySet()
        assay_set.add_assay(assay_x)
        for assay_id_y, overlap_rmsd in assay_x.overlaps.items():
            if overlap_rmsd <= 0.5:
                assay_set.add_assay(assays[assay_id_y])

        assay_sets.append(assay_set)
    return assay_sets


def build_sets_from_combinations(assays):
    """Combinatorial set building. Builds 10,000 at a time, then adds to ID_COMBINATIONS
    """
    from itertools import combinations
    assay_id_sets = {assay_id_set for n in range(len(assays)) for assay_id_set in combinations(assays.keys(), n + 1)}

    assay_sets = list()
    for assay_id_set in assay_id_sets:
        assay_set = AssaySet()
        assay_set.add_assays(*[assays[assay_id] for assay_id in assay_id_set])
        if len(assay_set) > 120:
            assay_sets.append(assay_set)
        if len(assay_sets) == 10000:
            add_to_global(assay_sets)
            assay_sets = list()


def extend_set(assay_sets, assays, cutoff=0.65, strict=True, method='extend_l'):
    """Expand our existing sets using assays which favourable overlap (if strict) or semi-randomly (but weighted
       towards better overlap).
    """
    id_combinations = {frozenset(assay_set.ids) for assay_set in assay_sets}

    new_sets = list()
    for assay_set in assay_sets:
        if method == 'append' or method == 'extend_l':
            current_set = assay_set
            assay_set = deepcopy(assay_set)

        potential_additions = assays.keys() - assay_set.ids

        for assay_id in potential_additions:
            random_cutoff = cutoff
            x_overlap = assays[assay_id].overlaps

            passes = int()
            for set_assay_id in assay_set.ids:
                overlap_rmsd = x_overlap.get(set_assay_id, None)
                if isinstance(overlap_rmsd, float):
                    if overlap_rmsd > 0.5:
                        break
                    passes += 1
            else:
                random_cutoff -= 0.4 if passes >= 3 else 0.2 if passes else 0
                if (strict is True and passes) or (strict is False and random() > random_cutoff):
                    assay_set.add_assay(assays[assay_id])
                    if method == 'append':
                        if assay_set.ids not in id_combinations:
                            new_sets.append(assay_set)
                        # noinspection PyUnboundLocalVariable
                        assay_set = deepcopy(current_set)
        if method != 'append' and assay_set.ids not in id_combinations:
            new_sets.append(assay_set)
    assay_sets += new_sets
    return assay_sets


def expand_with_random(assays, num_sets):
    """Totally random expansion, stratified across a range of set sizes in compounds. Builds in blocks of 10,000 and
    then adds these to ID_COMBINATIONS
    """
    assays_list = list(assays.values())

    def target_length(split): return split * int(MAX_LEN * 0.09 - 12) + 120

    for _ in range((num_sets - NUM_SETS) // 10000 + 4):  # Three extra attempts to make up for missing sets
        random_sets = list()

        sets_to_build = min(num_sets - NUM_SETS, 10000)
        if not sets_to_build > 0:
            break

        sets_per_split, remainder = divmod(sets_to_build, 10)
        for n in range(10):
            if n == 9:
                sets_per_split += remainder
            for y in range(sets_per_split):
                rand_target_length = randint(target_length(n), target_length(n + 1))
                assay_set = AssaySet()
                while len(assay_set) < rand_target_length and assay_set.ids < assays.keys():
                    assay_set.add_assay(choice(assays_list))
                random_sets.append(assay_set)
        add_to_global(random_sets)


def add_to_global(assay_sets):
    """Adds assay_sets to ID_COMBINATIONS.

    Assay sets are stored in memory/on disk as JSON serialized, UTF-8 encoded byte strings. ID_COMBINATIONS stores these
    in order of set size (in compounds), and LENGTH_BYTE_COUNTS keeps track of how many bytes of string belong to each
    set size. TEMP_ID_COMBINATIONS is used when adding new sets to the pile.

    If we know our set size is 239, we can seek to the number of bytes which are for set sizes fewer than 239
    (given by sum([LENGTH_BYTE_COUNTS[x] for x in range(1, 239)]) and check that our set's assay IDs aren't in the next
    LENGTH_BYTE_COUNTS[239] bytes. This is much faster than checking the whole ID_COMBINATIONS file.
    """
    global NUM_SETS, ID_COMBINATIONS, TEMP_ID_COMBINATIONS, LENGTH_BYTE_COUNTS
    ID_COMBINATIONS.seek(0)

    lengths_to_sets = dict()
    for assay_set in assay_sets:
        set_length = len(assay_set)
        if set_length < 120 or set_length > MAX_LEN:
            continue
        # Converting set to sorted, JSON formatted byte string.
        id_set_string = json.dumps(sorted(assay_set.ids)) + '\n'
        byte_string = bytes(id_set_string, encoding='utf-8')

        if set_length not in lengths_to_sets:
            lengths_to_sets[set_length] = [byte_string]
        else:
            if byte_string not in lengths_to_sets[set_length]:
                lengths_to_sets[set_length].append(byte_string)

    local_lengths = sorted(lengths_to_sets.keys())
    global_lengths = sorted(LENGTH_BYTE_COUNTS.keys())

    seek_positions = list()
    for length in local_lengths:
        # Set lengths pre this one.
        lengths_pre_insertion = global_lengths[:bisect_left(global_lengths, length)]
        start_cursor = sum([LENGTH_BYTE_COUNTS[global_length] for global_length in lengths_pre_insertion])
        seek_positions.append((length, start_cursor))

    for length, start_cursor in seek_positions:
        byte_strings = lengths_to_sets[length]

        # Writing all bytes between this length and the last in 1MiB blocks (.tell() gives current cursor position)
        mebibytes_to_copy, remaining_bytes = divmod(start_cursor - ID_COMBINATIONS.tell(), MEBIBYTE)
        for _ in range(mebibytes_to_copy):
            TEMP_ID_COMBINATIONS.write(ID_COMBINATIONS.read(MEBIBYTE))
        TEMP_ID_COMBINATIONS.write(ID_COMBINATIONS.read(remaining_bytes))

        bytes_to_check = LENGTH_BYTE_COUNTS.get(length, 0)
        if bytes_to_check:
            while True:
                # readlines reads to _end_ of last line (more bytes than requested)
                block_size = min(MEBIBYTE, (start_cursor + bytes_to_check) - ID_COMBINATIONS.tell())
                if not block_size:
                    break
                block = ID_COMBINATIONS.readlines(block_size)
                byte_strings = [byte_string for byte_string in byte_strings if byte_string not in block]
                TEMP_ID_COMBINATIONS.writelines(block)

        for byte_string in byte_strings:
            LENGTH_BYTE_COUNTS[length] += TEMP_ID_COMBINATIONS.write(byte_string)  # write() returns n bytes written
            NUM_SETS += 1

    while True:  # No longer care about position, just copying over the rest of the list
        block = ID_COMBINATIONS.read(MEBIBYTE)
        if block:
            TEMP_ID_COMBINATIONS.write(block)
        else:
            ID_COMBINATIONS.seek(0)
            ID_COMBINATIONS.truncate(0)
            break
    ID_COMBINATIONS, TEMP_ID_COMBINATIONS = TEMP_ID_COMBINATIONS, ID_COMBINATIONS


def build_model_sets(assays, num_sets, max_len):
    """Builds num_sets assay sets (from 120 to max_len compounds in size) from assays. Returns a generator which
    yields these sets as a list of assay IDs with the set number appended on the end.
    """
    global ID_COMBINATIONS, MAX_LEN
    MAX_LEN = max_len

    ID_COMBINATIONS.seek(0)
    ID_COMBINATIONS.truncate(0)

    n_combinations = 2**len(assays) - 1

    if n_combinations <= num_sets:
        build_sets_from_combinations(assays)
    else:
        assay_sets = build_sets(assays)
        # No randomness - these sets are expanded only if the overlap is favourable.
        assay_sets = extend_set(assay_sets, assays, strict=True, method='append')
        assay_sets = extend_set(assay_sets, assays, strict=True, method='extend_l')

        # The following expansion methods have an element of randomness, but will not add sets _known_ to have a
        # bad overlap with any set already present.
        assay_sets = extend_set(assay_sets, assays, strict=False, cutoff=0.9, method='extend_l')
        if len(assay_sets) < (num_sets // 10):
            assay_sets = extend_set(assay_sets, assays, strict=False, cutoff=0.8, method='append')
            assay_sets = extend_set(assay_sets, assays, strict=False, cutoff=0.8, method='extend_l')
        # Want 90% random assays to ensure we sample fairly evenly across the set sizes.
        n_sets_over_tenth = len(assay_sets) - (num_sets // 10)
        for _ in range(n_sets_over_tenth):
            assay_sets.pop(randint(0, len(assay_sets) - 1))
        add_to_global(assay_sets)

        expand_with_random(assays, num_sets)  # Totally random expansion.

    ID_COMBINATIONS.seek(0)
    assay_id_gen = (json.loads(byte_string) + [str(set_id)] for set_id, byte_string in enumerate(ID_COMBINATIONS, 1))

    return assay_id_gen, NUM_SETS
