from numpy import isnan
from statistics import mean, median

__docstring__ = """Classes for QSARSetBuilder, a program to help build a clean data set for QSAR modelling 
                   from raw public data. Designed to work with ChEMBL data.
                """
__author__ = 'Travis Hesketh (travis@optibrium.com)'
__copyright__ = 'Copyright (c) Optibrium Ltd 2018. All rights reserved.'
__version__ = '2018.03.1'
__licence__ = 'GNU GPL v3'  # https://www.gnu.org/licenses/gpl-3.0.en.html


class Assay:
    """
    Class which holds key information from each assay.

    Attributes:
        id: string - ASSAY_ID
        assay_type: string - TYPE
        description: string
        data: dict - SMILES string: P_VALUE float
        duplicates: dict - SMILES string: set of P_VALUE floats
        overlaps: dict - ASSAY_ID string for each overlapping Assay: RMSD float for overlap with that assay
        compounds: set - SMILES strings
        substrates: set
    """

    def __init__(self, assay_id, assay_type=None):
        self.id, self.assay_type, self.description = str(assay_id), str(assay_type), str()
        self.data, self.duplicates, self.overlaps = dict(), dict(), dict()
        self.compounds, self.substrates = set(), set()

    def __repr__(self):
        string = '{id}. Type: {assay_type}, Length: {0}, Substrates: {substrates}\nDescription: {description}'\
                 .format(len(self), **self.__dict__)
        return string

    #  <----- Dict operations -----> These treat the Assay as a dict (self.data)
    def __setitem__(self, compound, value):  # assay[SMILES] = value
        if compound not in self.data:
            self.data[compound] = value
            self.compounds.add(compound)
        else:
            if compound not in self.duplicates:
                self.duplicates[compound] = {self.data[compound], value}
            else:
                self.duplicates[compound].add(value)
            self.data[compound] = median(self.duplicates[compound])

    def __delitem__(self, compound):  # del assay[SMILES]
        del self.data[compound]
        if compound in self.duplicates:
            del self.duplicates[compound]
        self.compounds.remove(compound)

    def __getitem__(self, key): return self.data[key]  # value = assay[SMILES]
    def __contains__(self, key): return key in self.data  # if 'Oc1ccccc1' in assay
    def __iter__(self): return iter(self.data.items())  # for compound, value in assay:
    def __len__(self): return len(self.data)  # len(assay)

    # <----- Dict operations -----> These treat the Assay as a set of (SMILES, value) tuples): ==, !=, <, <=, >, >=
    def __eq__(self, other): return self.data.items() == other
    def __ne__(self, other): return self.data.items() != other
    def __lt__(self, other): return self.data.items() < other
    def __le__(self, other): return self.data.items() <= other
    def __gt__(self, other): return self.data.items() > other
    def __ge__(self, other): return self.data.items() >= other

    #  <----- Set operations -----> These treat the Assay as a set of SMILES strings:  -, &, |, ^
    def __sub__(self, other): return self.compounds - other
    def __rsub__(self, other): return other - self.compounds
    def __and__(self, other): return self.compounds & other
    def __rand__(self, other): return other & self.compounds
    def __or__(self, other): return self.compounds | other
    def __ror__(self, other): return other | self.compounds
    def __xor__(self, other): return self.compounds ^ other
    def __rxor__(self, other): return other ^ self.compounds

    def update_overlaps(self, assay, rmsd):
        assert isinstance(assay, Assay)
        self.overlaps[assay.id] = rmsd
        assay.overlaps[self.id] = rmsd


class AssaySet(Assay):
    """Class which holds the data from a set of assays, their split into test/training splits and the results.
        Attributes:
        <----- Inherited from Assay ----->
        id: string - AssaySet ID
        data: dict - SMILES string: P_VALUE float
        duplicates: dict - SMILES string: set of P_VALUE floats.
        compounds: set - SMILES strings.

        <----- AssaySet specific ----->
        assay_types: set - data types included in the AssaySet (e.g. {IC50, EC50})
        ids: set - assays included in the AssaySet
        trn_set: set - SMILES strings in the training set
        tst_set: set - SMILES strings in the test set
        descriptors: dict - SMILES string: list of floats (descriptor values)
        fingerprints: dict - SMILES string: list  of ints (Morgan fingerprint of path length two)
        r_descr_mask: list - integers. Indexes of the descriptors used in the regression model
        c_descr_mask: list - integers. Indexes of the descriptors used in the classification model
        r_results: list - tuples containing model_name string and rsq, rmse, val_rsq and val_rmse floats
        c_results: list - tuples containing model_name string and mcc, accu, val_mcc and val_accu floats
        split: int - the split used for the classification model
        results: string - a formatted string of the results
        good_set: bool - True if the set has Rsq/MCC and validated Rsq/MCC > 0.6 for any estimator
    """
    def __init__(self, set_id=None):
        super().__init__(set_id)
        del self.assay_type, self.description, self.overlaps, self.substrates  # Not needed in this class

        self.assay_types, self.ids, self.trn_set, self.tst_set = set(), set(), set(), set()
        self.descriptors, self.fingerprints = dict(), dict()
        self.r_descr_mask, self.c_descr_mask, self.r_results, self.c_results = list(), list(), list(), list()
        self.split = int()
        self.results = str()
        self.good_set = False

        #
    def __repr__(self):
        return '{id}. Length: {0}\nAssay Types: {assay_types}'.format(len(self.data), **self.__dict__)

    #  <----- Dict operations -----> These treat the AssaySet as a dict (self.data)
    def __delitem__(self, item): del self.data[item], self.descriptors[item], self.fingerprints[item]

    # <----- AssaySet methods ----->
    def add_assay(self, assay):
        if assay.id in self.ids:
            return

        self.ids.add(assay.id)
        self.assay_types.add(assay.assay_type)

        duplicate_compounds = self & assay

        for compound, value in assay:
            if compound not in duplicate_compounds:
                self[compound] = value
                self.compounds.add(compound)
            else:
                if compound in self.duplicates:  # If there are already duplicate values in the set...
                    if compound in assay.duplicates:  # ...and in the assay
                        self.duplicates[compound] |= assay.duplicates[compound]  # Set union of both sets of dupes.
                    else:  # ...but not in the assay
                        self.duplicates[compound].add(value)  # Add the assay's value to the set's duplicates
                else:  # If there are not duplicate values in the set...
                    if compound in assay.duplicates:  # ...but there are in the assay
                        self.duplicates[compound] = assay.duplicates[compound]  # Copy the assay's dupes to the set
                        self.duplicates[compound].add(self.data[compound])  # and add the assay's value
                    else:  # ...and not in the assay
                        self.duplicates[compound] = {self.data[compound], value}  # Add both values to the set's dupes
                # Data value to the mean of the duplicates with each new compound added.
                self.data[compound] = median(tuple(self.duplicates[compound]))

    def add_assays(self, *assays):
        for assay in assays:
            self.add_assay(assay)

    def format_results(self):
        if self.results:
            return
        header = 'Model run number: {id} Set size: {0} Types: {assay_types}\n'.format(len(self), **self.__dict__)
        true_false = 'Classification {0}: {1} / {2}\n'.format(*self.c_results.pop(0)) if self.c_results else str()
        output = header + true_false

        best_r_model, best_c_model = (str(), float()), (str(), float())
        for result_set in (self.r_results, self.c_results):
            for estimator_results in result_set:
                name, stat_1, stat_2, v_stat_1, v_stat_2 = estimator_results
                if not isnan(v_stat_1):
                    self.good_set = True
                    average_performance = mean((stat_1, v_stat_1))
                    if result_set is self.r_results and average_performance > best_r_model[1]:
                        best_r_model = (name, average_performance)
                    elif result_set is self.c_results and average_performance > best_c_model[1]:
                        best_c_model = (name, average_performance)

        if self.good_set:
            if best_r_model[1]:
                output += 'Good regression set. Best regression model: {}\n'.format(best_r_model[0])
            if best_c_model[1]:
                output += 'Good classification set. Best classification model: {}\n'.format(best_c_model[0])
        else:
            output += 'No tested models performed well.\n'
        self.results = output + '\n'


class TestSet:
    """Class which holds the data from the external test set.

    Attributes:
        data: dict - SMILES string: P_VALUE float
        compounds: set - SMILES strings.
        descriptors: dict - SMILES string: list of floats (descriptor values)
    """
    def __init__(self, assay_data, descriptor_dict):
        self.compounds = set()
        self.data, self.descriptors = dict(), dict()

        for datapoint in assay_data:
            compound = datapoint[2]
            value = datapoint[3]
            descrs = descriptor_dict[compound]

            self.data[compound] = value
            self.descriptors[compound] = descrs

        self.compounds = set(self.data.keys())

    def values(self):
        for compound, value in self.data.items():
            yield (self.descriptors[compound], value)
