import warnings
# numpy
import numpy
# rdkit imports
from rdkit.SimDivFilters.rdSimDivPickers import MaxMinPicker
# sklearn imports
from sklearn.preprocessing import RobustScaler
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.feature_selection import VarianceThreshold, RFE
from sklearn.model_selection import cross_val_predict
from sklearn.metrics import r2_score, mean_squared_error, accuracy_score, matthews_corrcoef

empty_array = numpy.ndarray([0])

__docstring__ = """Model building / testing for QSARSetBuilder, a program to help build a clean data set for QSAR 
                modelling from raw public data. Designed to work with ChEMBL data.
                """
__author__ = 'Travis Hesketh (travis@optibrium.com)'
__copyright__ = 'Copyright (c) Optibrium Ltd 2018. All rights reserved.'
__version__ = '2018.03.1'
__licence__ = 'GNU GPL v3'  # https://www.gnu.org/licenses/gpl-3.0.en.html

regressors = {'Random Forests': RandomForestRegressor(n_estimators=100)}

classifiers = {'Random Forests': RandomForestClassifier(n_estimators=100),
               'Decision Tree': DecisionTreeClassifier()}


def generate_scaling_function(assay_set):
    """Generates a scaling function from the assay's data and scikit-learn's RobustScaler."""
    return RobustScaler().fit(numpy.array(list(assay_set.descriptors.values()))).transform


def init_arrays(assay_set, scaling_function):
    """Converts assay_set into a list of strings (SMILES), a 1D array of floats (the P_VALUES), a 2D array of
    floats (descriptors) and a 2D array of floats (fingerprints). The descriptors are scaled using the scaling
    function."""
    compounds, values, descriptors, fingerprints = list(), list(), list(), list()
    for compound, value in assay_set:
        fingerprints.append(assay_set.fingerprints[compound])
        descriptors.append(assay_set.descriptors[compound])
        compounds.append(compound)
        values.append(value)

    value_array, descr_array = numpy.array(values), scaling_function(numpy.array(descriptors))

    return compounds, descr_array, value_array, fingerprints


def generate_descriptor_masks(descriptor_array, value_array, models_to_run, clsif_split):
    """Descriptor selection - returns two lists of integers, which are slices for the descriptor arrays."""
    r_descr_mask, c_descr_mask = None, None
    descriptor_mask = numpy.arange(descriptor_array.shape[1]).reshape(1, -1)

    var_thresh = VarianceThreshold(threshold=(0.9 * (1 - 0.90))).fit(descriptor_array)
    descriptor_array = var_thresh.transform(descriptor_array)
    descriptor_mask = var_thresh.transform(descriptor_mask)

    if 'r' in models_to_run:
        rfe_selector = RFE(RandomForestRegressor(), n_features_to_select=30).fit(descriptor_array, value_array)
        r_descr_mask = rfe_selector.transform(numpy.copy(descriptor_mask))[0].tolist()

    if 'c' in models_to_run:
        class_array = numpy.where(value_array >= clsif_split, numpy.int32(1), numpy.int32(0))
        rfe_selector = RFE(RandomForestClassifier(), n_features_to_select=30).fit(descriptor_array, class_array)
        c_descr_mask = rfe_selector.transform(numpy.copy(descriptor_mask))[0].tolist()

    return r_descr_mask, c_descr_mask


def apply_mask(mask, *arrays):
    """Applies a 1D integer index mask to a 2D array."""
    output_arrays = list()
    for array in arrays:
        if array is empty_array or not array.any():
            output_arrays.append(empty_array)
        else:
            output_arrays.append(array[:, mask])
    return output_arrays


def split_sets(compounds, fingerprints):
    """Splits compound fingerprints into test and training sets."""
    def tanimoto_similarity(x, y): return 1 - (numpy.bitwise_and(x, y) / numpy.bitwise_or(x, y))

    # noinspection PyArgumentList
    picker = MaxMinPicker()
    trn_set_indices = list(picker.LazyPick(tanimoto_similarity, len(fingerprints), int(0.7 * len(fingerprints))))
    tst_set_indices = [x for x in range(len(fingerprints)) if x not in trn_set_indices]

    trn_set = {compounds[x] for x in trn_set_indices}
    tst_set = {compounds[x] for x in tst_set_indices}
    return trn_set, tst_set, trn_set_indices, tst_set_indices


def score_clsif(tr_descrs, tr_classes, te_descrs, te_classes, e_t_descrs=empty_array, e_t_classes=empty_array):
    """Score classification models from the set."""
    clsif_results = list()

    tr_true, te_true = numpy.count_nonzero(tr_classes), numpy.count_nonzero(te_classes)
    tr_false, te_false = len(tr_classes) - tr_true, len(te_classes) - te_true

    # Cross validation fails otherwise.
    if not (tr_true > 5 and tr_false > 5) or not (te_true and te_false):
        return clsif_results  # Returning an empty list means that iterating over it doesn't fail.

    clsif_results.append(('True/False', tr_true + te_true, tr_false + te_false, numpy.nan, numpy.nan))
    for method_name, classifier in classifiers.items():
        classifier.fit(tr_descrs, tr_classes)
        predictions = classifier.predict(te_descrs)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')  # ZeroDivisionWarnings are messy and uninteresting
            try:
                mcc, accuracy = matthews_corrcoef(te_classes, predictions), accuracy_score(te_classes, predictions)
            except FloatingPointError:
                mcc, accuracy = numpy.nan, numpy.nan

            val_mcc, val_accuracy = numpy.nan, numpy.nan

            if mcc >= 0.6:  # Not validating if the model is poor.
                if e_t_descrs is not empty_array:
                    val_pred, val_obsd = classifier.predict(e_t_descrs), e_t_classes
                else:
                    val_pred, val_obsd = cross_val_predict(classifier, tr_descrs, tr_classes, cv=5), tr_classes

                val_mcc, val_accuracy = matthews_corrcoef(val_obsd, val_pred), accuracy_score(val_obsd, val_pred)
                if val_mcc < 0.6:
                    val_mcc, val_accuracy = numpy.nan, numpy.nan

            kappa, accuracy, val_kappa, val_accuracy = [round(x, 2) for x in [mcc, accuracy, val_mcc, val_accuracy]]
            clsif_results.append((method_name, kappa, accuracy, val_kappa, val_accuracy))

    return clsif_results


def score_rgrsn(tr_descrs, tr_values, te_descrs, te_values, e_t_descrs=empty_array, e_t_values=empty_array):
    """Score regression models from the set."""
    rgrsn_results = list()
    for method_name, regressor in regressors.items():
        regressor.fit(tr_descrs, tr_values)
        predictions = regressor.predict(te_descrs)
        r_sq, rmse = r2_score(te_values, predictions), numpy.sqrt(mean_squared_error(te_values, predictions))

        val_r_sq, val_rmse = numpy.nan, numpy.nan
        if r_sq > 0.6:  # Not validating if the model is poor.
            if e_t_descrs is not empty_array:
                val_pred, val_obsd = regressor.predict(e_t_descrs), e_t_values
            else:
                val_pred, val_obsd = cross_val_predict(regressor, tr_descrs, tr_values, cv=5), tr_values

            val_r_sq, val_rmse = r2_score(val_obsd, val_pred), numpy.sqrt(mean_squared_error(val_obsd, val_pred))
            if val_r_sq < 0.6:
                val_r_sq, val_rmse = numpy.nan, numpy.nan

        r_sq, rmse, val_r_sq, val_rmse = [round(x, 2) for x in [r_sq, rmse, val_r_sq, val_rmse]]
        rgrsn_results.append((method_name, r_sq, rmse, val_r_sq, val_rmse))

    return rgrsn_results


def fit_and_test_models(assay_set, clsif_split, models_to_run, ext_test_set=None):
    """Select descriptors, split sets into training and validation sets and build/test models. If the model scores
    > 0.6 (Rsq or MCC), cross validate using 5 fold cross validation or an external test set from file."""
    e_t_descr_array, e_t_value_array, e_t_class_array = empty_array, empty_array, empty_array
    scaling_func = generate_scaling_function(assay_set)

    if ext_test_set:
        for compound in assay_set.compounds & ext_test_set.compounds:
            del assay_set[compound]
        e_t_descrs, e_t_values = zip(*ext_test_set.values())
        e_t_descr_array, e_t_value_array = numpy.array(e_t_descrs), numpy.array(e_t_values)
        e_t_class_array = numpy.where(e_t_value_array >= clsif_split, numpy.int32(1), numpy.int32(0))
        e_t_descr_array = scaling_func(e_t_descr_array)

    compounds, descr_array, value_array, fingerprints = init_arrays(assay_set, scaling_func)
    class_array = numpy.where(value_array >= clsif_split, numpy.int32(1), numpy.int32(0))

    r_descr_mask, c_descr_mask = generate_descriptor_masks(descr_array, value_array, models_to_run, clsif_split)
    assay_set.trn_set, assay_set.tst_set, trn_set_indices, tst_set_indices = split_sets(compounds, fingerprints)

    if 'r' in models_to_run:
        r_descrs, e_r_descrs = apply_mask(r_descr_mask, descr_array, e_t_descr_array)
        r_tr_descrs, tr_values = zip(*[(r_descrs[x], value_array[x]) for x in trn_set_indices])
        r_te_descrs, te_values = zip(*[(r_descrs[x], value_array[x]) for x in tst_set_indices])

        rgrsn_results = score_rgrsn(r_tr_descrs, tr_values, r_te_descrs, te_values, e_r_descrs, e_t_value_array)
        assay_set.r_results, assay_set.r_descr_mask = rgrsn_results, r_descr_mask

    if 'c' in models_to_run:
        c_descrs, e_c_descrs = apply_mask(c_descr_mask, descr_array, e_t_descr_array)
        c_tr_descrs, tr_classes = zip(*[(c_descrs[x], class_array[x]) for x in trn_set_indices])
        c_te_descrs, te_classes = zip(*[(c_descrs[x], class_array[x]) for x in tst_set_indices])

        clsif_results = score_clsif(c_tr_descrs, tr_classes, c_te_descrs, te_classes, e_c_descrs, e_t_class_array)
        assay_set.c_results, assay_set.c_descr_mask = clsif_results, c_descr_mask

    assay_set.format_results()
    return assay_set
