import pickle
import gzip
import os
import errno
import sys
from hashlib import sha1

__docstring__ = """Utility funcitons for QSARSetBuilder, a program to help build a clean data set for QSAR modelling 
                   from raw public data. Designed to work with ChEMBL data.
                """
__author__ = 'Travis Hesketh (travis@optibrium.com)'
__copyright__ = 'Copyright (c) Optibrium Ltd 2018. All rights reserved.'
__version__ = '2018.03.1'
__licence__ = 'GNU GPL v3'  # https://www.gnu.org/licenses/gpl-3.0.en.html


def load_pickle(file_path):
    """Loads data from a gzipped pickle."""
    with gzip.open(file_path, 'rb') as file_object:
        pickled_data = pickle.load(file_object)
        return pickled_data


def save_pickle(data_to_pickle, file_path):
    """Saves data to a gzipped pickle."""
    with gzip.open(file_path, 'wb') as file_object:
        pickle.dump(data_to_pickle, file_object)


def create_directory(directory):
    """Creates the directory specified, or, if that's not possible, creates it in the working directory and returns
    the path."""
    try:
        os.makedirs(directory)
    except OSError as err:
        if err.errno != errno.EEXIST:
            directory_in_cwd = os.path.join(os.getcwd(), os.path.split(directory)[1])
            if directory != directory_in_cwd:
                directory = directory_in_cwd
                sys.stdout.write('Failed to create directory in specified location. Creating directory in ' + directory
                                 + 'instead.')
                create_directory(directory)
            else:
                raise IOError('Could not create directory ' + directory)
    return directory


def get_hash(file):
    """Calculates the SHA1 hash for a file and returns the hex digest."""
    with open(file, 'rb', buffering=0) as input_file_object:
        sha1_hash = sha1()
        while True:
            block = input_file_object.read(128 * 1024)  # Reading 128kB at a time.
            if not block:
                break
            sha1_hash.update(block)
        return sha1_hash.hexdigest()


def time_diff(initial_time, end_time):
    """Given a start and end time in seconds, returns the number of hours, minutes and seconds."""
    m, s = divmod(end_time - initial_time, 60)
    h, m = divmod(m, 60)
    return [str(round(x)).rjust(2, '0') for x in (h, m, s)]


def set_process_priority(priority):
    """Sets priority of the process on Windows and UNIX systems. May work on others if os.nice is implemented."""
    if os.name == 'nt':
        magic_priority_values = dict(idle=0x0040, below_normal=0x4000, normal=0x0020, above_normal=0x8000, high=0x0080)
        from ctypes import WinDLL, wintypes
        kernel32 = WinDLL('kernel32')
        priority_value = wintypes.DWORD(magic_priority_values[priority])
        kernel32.SetPriorityClass(kernel32.GetCurrentProcess(), priority_value)
    else:
        # Root access required to increase priority on unix systems - negative values won't work.
        magic_priority_values = dict(idle=5, below_normal=2, normal=0, above_normal=0, high=0)
        from os import nice
        if nice(0) == 0:
            os.nice(magic_priority_values[priority])
